; ColorPRINT
; version 1.0
; Copyright (c) 2008 Andrew Owen

; -----------------------------------------------------------------------------
; ColorPRINT is a MODE 2 characater printing tool for the Timex TC2048 and
; Spectrum SE computers. It uses channels/streams to extend 48 or 128 BASIC.
; Use it with PRINT #4 CHR$ n (where n is 0 to 255). The co-ordinates are taken
; from the standard screen channel. To print at a specific co-ordiante use
; PRINT AT y,x; #4; "some text";. All 256 characters can be redefined. USR "A"
; is the first byte of the bitmap of character 144. USR "A" + 2048 is the first
; byte of the attributes of character 144. In the font area, USR "A" - 376 is
; the start of the bitmap of 'a', while USR "A" - 632 is 'A' and so on (the
; CHARS and UDG system variables are changed to point at the new data area).
; PRINT, PLOT, CIRCLE and DRAW now function on the bitmap only and will not
; modify the attributes.

; -----------------------------------------------------------------------------
; MEMORY MAP
;
;     +---------+--------+---------+-----------+------------...
;     | Display | Do not | Printer | System    | Microdrive
;     | File    | use    | Buffer* | variables | maps
;     +---------+--------+---------+-----------+------------...
;     |         |        |         |           |
;   16384     22528    23296     23552       23734
;
; ...-+------------+-------------+---------+-----------+-...
;     | Attributes | Channel     | BASIC   | Variables | etc.
;     |            | information | program |           |
; ...-+------------+-------------+---------+-----------+-...
;     |            |             |         |           |
;   24576        30720          PROG      VARS       E_LINE
;
; ...-+------+-------+--------+----------+------+---------+------------+
;     | 0x3E | ASCII | Normal | Block    | UDGs | ASCII   | Attributes |
;     |      | 0-31  | font   | graphics | A-U  | 165-255 | 0-255      |
; ...-+------+----------------+----------+------+---------+------------+
;     |      |       |        |          |      |         |            |
;  RAMTOP  61440   61696    62464      62592  62760     63488        65535
;
; * Extra system variables in 128 BASIC.

; -----------------------------------------------------------------------------
; ROM ROUTINES, SYSTEM VARIABLES and CONSTANTS

       pr_all          equ     0x0b7f  ; print all routine
       po_store        equ     0x0adc  ; position store routine
       report_j        equ     0x15c4  ; null input routine
       make_room       equ     0x1655  ; create buffer routine
       strms           equ     0x5c10  ; pointers to individual streams
       chars           equ     0x5c36  ; start of font - 256
       chans           equ     0x5c4f  ; start of channel area
       prog            equ     0x5c53  ; start of BASIC area
       udg             equ     0x5c7b  ; start of UDGs
       df_cc           equ     0x5c84  ; display file addres
       s_posn          equ     0x5c88  ; line/column
       attr            equ     24576   ; start of attributes

; -----------------------------------------------------------------------------
; START

       org     61264

; -----------------------------------------------------------------------------
; SAVE REGISTERS

       push    af
       push    bc
       push    de
       push    hl

; -----------------------------------------------------------------------------
; MAIN

       call    move_chans
       call    make_chan
       call    set_chars_udg
       ld      a, %01000111            ; INK 7, PAPER 0, BRIGHT 1, FLASH 0
       call    cls

; -----------------------------------------------------------------------------
; RESTORE REGISTERS

       pop     hl
       pop     de
       pop     bc
       pop     af

; -----------------------------------------------------------------------------
; EXIT TO BASIC

       ret                             ; return

; -----------------------------------------------------------------------------
; COLOR PRINT CHANNEL

c_print:
        ld        bc, (chars)          ; font address
        ld        h, 0                 ; set high byte to 0
        ld        l, a                 ; character to L
        add       hl, hl               ; multiply by 8 to point to first byte
        add       hl, hl               ;
        add       hl, hl               ;
        add       hl, bc               ;
        ex        de, hl               ; character address to DE
        ld        bc, (s_posn)         ; get line/column
        ld        hl, (df_cc)          ; get display file address
        push      de                   ; store character address
        call      pr_all               ; pr-all
        call      po_store             ; po-store
        pop       de                   ; get character address
        ex        de, hl               ; store it in HL
        ld        bc, 2048             ; offset to character attributes
        add       hl, bc               ; add offset
        ex        de, hl               ; character attribute address to DE
        ld        bc, 8192             ; offset to screen attributes
        add       hl, bc               ; add offset
        dec       hl                   ; back one position
        ld        c, 8                 ; loop counter
loop:
        ld        a, (de)              ; get the value
        ld        (hl), a              ; write it
        inc       h                    ; next screen position
        inc       de                   ; next attribute byte
        dec       c                    ; next counter
        jr        nz, loop             ; finished?
        ret                            ; return

; -----------------------------------------------------------------------------
; MOVE CHANNELS

move_chans:
       ld      hl, 0x7800              ; first address after attributes
       ld      de,(chans)              ; current start of channel data
       sbc     hl, de                  ; calculate how much to add
       ld      b, h                    ; transfer result to BC
       ld      c, l                    ;
       ex      de, hl                  ; start of channel data to HL
       dec     hl                      ; address previous byte
       call    make_room               ; make the space
       ret

; -----------------------------------------------------------------------------
; MAKE CHANNEL and ATTACH STREAM
;
; Based on code by Ian Beardsmore from Your Spectrum #7, September 1984.

make_chan:
       ld      hl, (prog)              ; a channel must be created below BASIC
                                       ; so look at the system variable PROG
       dec     hl                      ; move HL down one address
       ld      bc, 0x0005              ; the new channel takes 5 bytes
       call    make_room               ; call the MAKE_ROOM routine
       inc     hl                      ; move HL up one address
       ld      bc, c_print             ; could write the bytes directly but
                                       ; then code would be non-relocatable
       ld      (hl), c                 ; low byte of the output routine
       inc     hl                      ; move HL up one address
       push    hl                      ; save this address for later
       ld      (hl), b                 ; high byte of the output routine
       inc     hl                      ; move HL up one address
       ld      bc, report_j            ; address of input routine
       ld      (hl), c                 ; low byte of the input routine
       inc     hl                      ; move HL up one address
       ld      (hl), b                 ; high byte of the input routine
       inc     hl                      ; move HL up one address
       ld      (hl), 'S'               ; channel type: 'S' = screen

; attach stream
       pop     hl                      ; the first address plus one of the
                                       ; extra space stored earlier
       ld      de, (chans)             ; store the contents of CHANS in DE
       and     a                       ; clear the carry flag before
                                       ; calculation
       sbc     hl,de                   ; the difference between the start of
                                       ; the channels area and the start of the
                                       ; extra space becomes the offset, stored
                                       ; in HL
       ex      de, hl                  ; store the offset in DE
       ld      hl, strms               ; store the contents of STRMS in HL
       ld      a, 4                    ; stream number #4
       add     a, 3                    ; take account of streams -3 to -1
       add     a, a                    ; each of the seven default streams has
                                       ; two bytes of offset data
                                       ; the total number of bytes occupied,
                                       ; held in a, forms the offset for the
                                       ; new stream
       ld      b, 0x00                 ; set b to hold 0x00
       ld      c,a                     ; set the low byte of the offset
       add     hl,bc                   ; the offset is added to the base
                                       ; address to give the correct location
                                       ; in the streams table to store the
                                       ; offset
       ld      (hl),e                  ; the low byte of the offset
       inc     hl                      ; move HL up one address
       ld      (hl),d                  ; the high byte of the offset
       ret                             ; done

; -----------------------------------------------------------------------------
; SET CHARS and UDG

set_chars_udg:
       ld      hl, chars               ; address system variable
       ld      (hl), 0x00              ; set
       inc     hl                      ; font
       ld      (hl), 0xf0              ; address
       ld      hl, udg                 ; address system variable
       ld      (hl), 0x80              ; set
       inc     hl                      ; udg
       ld      (hl), 0xf4              ; address
       ret                             ; done

; -----------------------------------------------------------------------------
; CLS

cls:
       push    af                      ; save attribute value
       call    0x0d6b                  ; clear bitmap
       pop     af                      ; restore attirbute value
       ld      hl, attr                ; set attributes to value in A
       ld      de, attr + 1            ;
       ld      (hl), a                 ; set first byte
       ld      bc, 0x17ff              ;
       ldir                            ; set remaining 6133 bytes
       ret                             ; done

; -----------------------------------------------------------------------------
; FONT/UDG/ATTRIBUTES

       incbin  "ColorFONT.bin"
